/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */
package org.ow2.petals.se.jsr181;

import java.util.Iterator;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.activation.DataHandler;
import javax.jbi.messaging.MessagingException;
import javax.xml.transform.Source;

import org.apache.axiom.attachments.Attachments;
import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.soap.SOAPBody;
import org.apache.axiom.soap.SOAPEnvelope;
import org.apache.axiom.soap.SOAPFactory;
import org.apache.axis2.AxisFault;
import org.apache.axis2.context.ConfigurationContext;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.description.AxisOperation;
import org.apache.axis2.description.AxisService;
import org.apache.axis2.description.TransportOutDescription;
import org.apache.axis2.engine.AxisEngine;
import org.apache.axis2.wsdl.WSDLConstants;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.se.jsr181.spi.JBIContext;
import org.ow2.petals.se.jsr181.spi.JBIContextManager;
import org.ow2.petals.se.jsr181.spi.MessageSender;

/**
 * The JBIListener used to handle incoming JBI messages. This listener will
 * invoke the linked Axis2 service.
 * 
 * @author Christophe HAMERLING - eBM WebSourcing
 * @since 1.0
 * 
 */
public class JBIListener extends AbstractJBIListener implements MessageSender {

    private Logger logger;

    /**
     * 
     */
    public JBIListener() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.listener.AbstractListener#init()
     */
    @Override
    public void init() {
        this.logger = this.getLogger();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.listener.AbstractJBIListener#onJBIMessage(org.ow2.petals.component.framework.api.message.Exchange)
     */
    @Override
    public boolean onJBIMessage(Exchange exchange) {
        if (!exchange.isProviderRole()) {
            logger.info("Role not supported in JSR181 JBIListener : " + exchange.getRole());
            return false;
        }

        // get service
        AxisService axisService = this.getAxisService();
        if (axisService == null) {
            String message = "Can not find the AxisService for the current JBIListener";
            this.handleError(message, exchange);
            return false;
        }

        // check if the operation is compatible
        if (!checkOperation(exchange, axisService)) {
            String message = "Can not find the operation " + exchange.getOperation()
                    + " in AxisService";
            this.handleError(message, exchange);
            return false;
        }

        // check if the MEP is compatible
        if (!checkMEP(exchange, axisService)) {
            String message = "MEP " + exchange.getPattern()
                    + " is not compatible with AxisService operation MEP";
            this.handleError(message, exchange);
            return false;
        }

        // OK let's go...
        MessageContext inMessageContext = this.getAxisConfigurationContext().createMessageContext();
        ClassLoader save = Thread.currentThread().getContextClassLoader();
        // set the sender in order to be able to get it in the POJO class
        JBIContextManager.setJBIContext(new JBIContext(exchange, this));
        try {

            this.fillInAxisMessageContext(inMessageContext, axisService, exchange);
            this.fillInEnvelope(inMessageContext, exchange);

            AxisEngine.receive(inMessageContext);

            // need to switch Classloader for JAXBMarshaller
            Thread.currentThread().setContextClassLoader(axisService.getClassLoader());

            if (exchange.isInOutPattern() || exchange.isInOptionalOutPattern()) {
                MessageContext outMessageContext = inMessageContext.getOperationContext()
                        .getMessageContext(
                                org.apache.axis2.wsdl.WSDLConstants.MESSAGE_LABEL_OUT_VALUE);
                outMessageContext.setProperty(
                        org.apache.axis2.jaxws.spi.Constants.CACHE_CLASSLOADER, axisService
                                .getClassLoader());
                this.fillOutJBIMessage(outMessageContext, exchange);
            } else {
                // ...
            }

        } catch (AxisFault axisFault) {
            this.handleFault(axisFault, exchange);
        } catch (MessagingException me) {
            this.handleFault(me, exchange);
        } finally {
            JBIContextManager.setJBIContext(null);
            Thread.currentThread().setContextClassLoader(save);
        }
        return true;
    }

    /**
     * 
     * @return
     */
    protected AxisService getAxisService() {
        return getJSR181Component().getJSR181Context().getService(getProvides());
    }

    /**
     * Check if the MEP of the JBI message exchange and the Axis Operation are
     * the same.
     * 
     * @param exchange
     * @param axisService
     * @throws Exception
     */
    protected boolean checkMEP(Exchange exchange, AxisService axisService) {
        AxisOperation operation = axisService.getOperation(exchange.getOperation());
        if (this.logger.isLoggable(Level.FINE)) {
            logger.fine("MEP for Axis Operation is " + operation.getMessageExchangePattern());
            logger.fine("MEP for JBI message is " + exchange.getPattern().toString());
        }

        int mepConstant = operation.getAxisSpecificMEPConstant();

        return (exchange.isInOnlyPattern() && mepConstant == WSDLConstants.MEP_CONSTANT_IN_ONLY)
                || (exchange.isInOptionalOutPattern() && mepConstant == WSDLConstants.MEP_CONSTANT_IN_OPTIONAL_OUT)
                || (exchange.isInOutPattern() && mepConstant == WSDLConstants.MEP_CONSTANT_IN_OUT)
                || (exchange.isRobustInOnlyPattern() && mepConstant == WSDLConstants.MEP_CONSTANT_ROBUST_IN_ONLY);
    }

    /**
     * Check if an operation with the JBI exchange operation name is available
     * in the Axis Service.
     * 
     * @param exchange
     * @param axisService
     * @throws Exception
     */
    protected boolean checkOperation(Exchange exchange, AxisService axisService) {
        AxisOperation op = axisService.getOperation(exchange.getOperation());
        return op != null;
    }

    /**
     * 
     * @param messageContext
     * @param exchange
     */
    private void fillInAxisMessageContext(MessageContext messageContext, AxisService axisService,
            Exchange exchange) throws MessagingException {
        messageContext.setServerSide(true);
        messageContext.setAxisService(axisService);
        TransportOutDescription out = new TransportOutDescription("JBI");
        out.setSender(new EmptyTransportSender());
        messageContext.setTransportOut(out);

        // map jbi operation to soap action
        messageContext.setSoapAction(exchange.getOperationName());

        // attachments
        Set<DataHandler> jbiAttachments = exchange.getInMessageAttachments();
        Attachments attachments = new Attachments();
        Iterator<DataHandler> iter = jbiAttachments.iterator();
        while (iter.hasNext()) {
            DataHandler dataHandler = (DataHandler) iter.next();
            attachments.addDataHandler(dataHandler.getName(), dataHandler);
        }
        messageContext.setAttachmentMap(attachments);

        // TODO
    }

    /**
     * 
     * @param messageContext
     * @param exchange
     */
    private void fillOutJBIMessage(MessageContext messageContext, Exchange exchange) {
        SOAPEnvelope response = messageContext.getEnvelope();
        if (logger.isLoggable(Level.FINE)) {
            logger.fine("RESPONSE : " + response);
        }

        try {
            Source source = AxiomHelper.createSource(response);
            if (source != null) {
                exchange.setOutMessageContent(source);
            }
        } catch (MessagingException e) {
            // TODO : Set an error on MEX
            this.logger.log(Level.WARNING, "Can not set message response", e);
        } catch (Exception e) {
            // TODO : Set an error on MEX
            this.logger.log(Level.WARNING, "Can not set message response", e);
        }
    }

    /**
     * 
     * @param messageContext
     * @param exchange
     * @return
     */
    private void fillInEnvelope(MessageContext messageContext, Exchange exchange)
            throws MessagingException, AxisFault {
        // TODO get good factory
        SOAPFactory factory = OMAbstractFactory.getSOAP11Factory();
        SOAPEnvelope envelope = factory.createSOAPEnvelope();
        SOAPBody body = AxiomHelper.fillSOAPBodyFromSource(factory, envelope, exchange
                .getInMessage().getContent());
        envelope.addChild(body);
        messageContext.setEnvelope(envelope);
    }

    /**
     * 
     * @return
     */
    private Component getJSR181Component() {
        return (Component) getComponent();
    }

    /**
     * 
     * @return
     */
    private ConfigurationContext getAxisConfigurationContext() {
        return getJSR181Component().getAxisContext();
    }

    /**
     * Handle a fault
     * 
     * @param message
     * @param exchange
     */
    private void handleFault(String message, Exchange exchange) {
        this.handleFault(new Exception(message), exchange);
    }

    /**
     * Handle a fault
     * 
     * @param exception
     * @param exchange
     */
    private void handleFault(Exception exception, Exchange exchange) {
        logger.warning("FAULT : " + exception.getMessage());
        if (exchange.isInOnlyPattern()) {
            // want to set the fault !!!
            exchange.setError(new Exception("The jsr-181 service execution generated an exception", exception));
        } else {
            try {
                exchange.setFault(exception);
            } catch (MessagingException e) {
                this.logger.log(Level.WARNING, "Can not set the fault on message exchange");
                exchange.setError(new Exception("The jsr-181 service generated an exception and cannot manage it", e));
            }
        }
    }

    /**
     * Handle an error
     * 
     * @param message
     * @param exchange
     */
    private void handleError(String message, Exchange exchange) {
        this.handleError(new Exception(message), exchange);
    }

    /**
     * Handle an error
     * 
     * @param exception
     * @param Exchange
     */
    private void handleError(Exception exception, Exchange exchange) {
        logger.warning("ERROR : " + exception.getMessage());
        exchange.setError(new Exception("The jsr-181 service execution generated an exception", exception));
    }
}
